#ifndef __JABYENGINE_SPU_PORTS_HPP__
#define __JABYENGINE_SPU_PORTS_HPP__
#include "../Auxiliary/io_class_helper.hpp"
#include "../../limits.h"

namespace SPU {
    enum Mode {
        Linear      = 0,
        Exponential = 1,
    };
    
    struct __no_align SampleRate {
        uint16_t raw_value = 0;

        static constexpr SampleRate from_HZ(long double freq)
        {
            //SampleRate is defined as 4096 == 44100Hz
            constexpr long double Base = (4096.0 / 44100.0);

            return static_cast<SampleRate>(((freq*4096.0)/44100.0));
        }
    };

    struct __no_align SweepVolume {
        typedef int16_t VolumeStep;

        static constexpr VolumeStep VolumeMax        = 1000;
        static constexpr VolumeStep SingleVolumeStep = (I16_MAX >> 1)/VolumeMax; 
        
        enum Direction {
            Increase = 0,
            Decrease = 1,
        };

        enum Phase {
            Positive = 0,
            Negative = 1,
        };

        int16_t raw_value = 0;

        constexpr SweepVolume() = default;

        constexpr auto& set_volume_percent(double percent) {
            this->raw_value = bit::value::set_normalized(this->raw_value, static_cast<int16_t>((I16_MAX/100.0)*percent), __start_end_bit2_start_length(0, 14));
            return *this;
        }

        //A value between 0 and 1000
        constexpr auto& set_volume_step(VolumeStep volume_step) {
            volume_step *= SingleVolumeStep;

            this->raw_value = bit::value::set_normalized(this->raw_value, volume_step, __start_end_bit2_start_length(0, 14));
            return *this;
        }

        //A value between 0 and 1000
        constexpr VolumeStep get_volume_step() const {
            return (bit::value::get_normalized(this->raw_value, __start_end_bit2_start_length(0, 14))/SingleVolumeStep);
        }

        io_class__2option_map(volume_mode, sweep_mode, 15);
        io_class__2option_map_getter(Mode,      linear_sweep_mode,    exponential_sweep_mode, sweep_mode_type,      14);
        io_class__2option_map_getter(Direction, increase_sweep_mode,  decrease_sweep_mode,    sweep_mode_direction, 13);
        io_class__2option_map_getter(Phase,     positive_sweep_phase, negative_sweep_phase,   sweep_phase,          12);

        //Uses only 5bit of shift (0..1F (slow..fast))
        constexpr auto& set_sweep_shift(uint8_5b shift) {
            const int16_t crop_value = static_cast<int16_t>(shift); //(bit::value::crop_value(shift, 5));
            this->raw_value = bit::value::set_normalized(this->raw_value, crop_value, __start_end_bit2_start_length(2, 6));
            return *this;
        }

        constexpr uint8_5b get_sweep_shift() const {
            return bit::value::get_normalized(this->raw_value, __start_end_bit2_start_length(2, 6));
        }
        
        //0..3 maps to => +7, +6, +5, +4 or -8, -7, -6, -5
        constexpr auto& set_sweep_step(uint8_2b step) {
            this->raw_value = bit::value::set_normalized(this->raw_value, static_cast<int16_t>(step), __start_end_bit2_start_length(0, 1));
            return *this;
        }

        constexpr uint8_2b get_sweep_step() const {
            return bit::value::get_normalized(this->raw_value, __start_end_bit2_start_length(0, 1));
        }
    };

    struct __no_align ADSR {
        uint32_t raw_value = 0;
    };

    struct __no_align Voice {
        static constexpr                 size_t Count   = 24;
        static inline    __always_inline auto&  Channel = reinterpret_cast<Voice(&)[Count]>(*reinterpret_cast<Voice*>(0x1f801c00));

        SweepVolume volumeLeft;
        SweepVolume volumeRight;
        SampleRate  sampleRate;
        uint16_t    adr;
        ADSR        adsr;
        SweepVolume currentVolume; //Not used
        uint16_t    repeatAdr;
    };

    struct MainVolume {
        static inline __always_inline auto& Left  = *reinterpret_cast<SweepVolume*>(0x1F801D80);
        static inline __always_inline auto& Right = *reinterpret_cast<SweepVolume*>(0x1F801D82);
    };
}

#endif //!__JABYENGINE_SPU_PORTS_HPP__